/*
 * config.c --- Parse a configuration file and build a key,value structure.
 * 14-may-2004/FK Dropped rotsim.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <unistd.h>
#include <assert.h>

#ifndef CFG_MAIN

#include "main.h"
#include "tonegen.h"
#include "scanner.h"
#include "midi.h"
#include "reverb.h"
#include "whirl.h"

#endif /* CFG_MAIN */

#include "config.h"

#define LINEBUFSZ 2048

/*
 * Each configurable module implements this function. The implementation
 * is idempotent. The most recent call defines the parameter's value.
 */
static int distributeParameter (ConfigContext * cfg) {

  int n = 0;

#ifdef CFG_MAIN

  printf ("%s:%d:[%s]=[%s]\n",
	  cfg->fname,
	  cfg->linenr,
	  cfg->name,
 	  cfg->value);

#else

  n += genConfig (cfg);
  n += midiConfig (cfg);
  n += pgmConfig (cfg);
  n += oscConfig (cfg);
  n += scannerConfig (cfg);
  n += ampConfig (cfg);
  n += whirlConfig (cfg);
  n += reverbConfig (cfg);

  if (n == 0) {
    printf ("%s:%d:%s=%s:Not claimed by any module.\n",
	    cfg->fname,
	    cfg->linenr,
	    cfg->name,
	    cfg->value);
  }

#endif /* CFG_MAIN */

  return n;
}

/*
 *
 */
int parseConfigurationLine (char * fname, int lineNumber, char * oneLine) {
  char delim[] = "=\n";
  char * s = oneLine;
  char * name;
  char * value;

  while (isspace (*s)) s++;	/* Skip over leading spaces */
  if (*s == '\0') return 0;	/* Skip empty lines */
  if (*s == '#') return 0;	/* Skip comment lines */
  if (*s == '=') {
    fprintf (stderr,
	     "%s:line %d: empty parameter name.\n",
	     fname,
	     lineNumber);
    return 0;
  }

  if ((name = strtok (s, delim)) != NULL) {
    int i;

	for (i = strlen (name) - 1; isspace (name[i]); name[i] = '\0', i--);

	if ((value = strtok (NULL, delim)) != NULL) {
	  char * t;
	  while (isspace (value[0])) value++;
	  for (t = value; *t != '\0'; t++) {
	    if (*t == '#') {	/* Terminate value at comment */
	      *t = '\0';
	      break;
	    }
	  }
	}
	else {
	  value = "";
	}

	i = strlen (value);
	if (0 < i) {
	  for (i = i - 1;  isspace (value[i]); value[i] = '\0', i--);
	}

	if (strcasecmp (name, "config.read") == 0) {
	  parseConfigurationFile (value);
	}
	else {
	  ConfigContext cfg;
	  cfg.fname  = fname;
	  cfg.linenr = lineNumber;
	  cfg.name   = name;
	  cfg.value  = value;
	  distributeParameter (& cfg);
	}
      }

}

/*
 *
 */
int parseConfigurationFile (char * fname) {
  int lineNumber = 0;
  char lineBuf [LINEBUFSZ];
  FILE * fp;

  if ((fp = fopen (fname, "r")) == NULL) {
    perror (fname);
    return -1;
  }
  else {

    while (fgets (lineBuf, LINEBUFSZ, fp) != NULL) {
      lineNumber += 1;		/* Increment the linenumber. */
      parseConfigurationLine (fname, lineNumber, lineBuf);
    }

    fclose (fp);
  }

  return 0;
}

/*
 *
 */
void showConfigfileContext (ConfigContext * cfg, char * msg) {

  assert (cfg != NULL);
  assert (msg != NULL);

  fprintf (stderr,
	   "%s:line %d:name %s:value %s:%s\n",
	   cfg->fname,
	   cfg->linenr,
	   cfg->name,
	   cfg->value,
	   msg);
}

/*
 *
 */
void configIntUnparsable (ConfigContext * cfg) {
  assert (cfg != NULL);
  showConfigfileContext (cfg, "value is not an integer");
}

/*
 *
 */
void configDoubleUnparsable (ConfigContext * cfg) {
  assert (cfg != NULL);
  showConfigfileContext (cfg, "value is not a real");
}

/*
 *
 */
void configDoubleOutOfRange (ConfigContext * cfg,
			     double lowest,
			     double highest) {
  char buf[256];
  assert (cfg != NULL);
  sprintf (buf, "real value out of range (%lf -- %lf)", lowest, highest);
  showConfigfileContext (cfg, buf);
}

/*
 *
 */
void configIntOutOfRange (ConfigContext * cfg, int lowest, int highest) {
  char buf[256];
  assert (cfg != NULL);
  sprintf (buf, "integer value out of range (%d -- %d)", lowest, highest);
  showConfigfileContext (cfg, buf);
}

/*
 * Assigns an sample range integer from the configuration context's value.
 * If the value is real (contains a period), it is interpreted as a
 * normalised quantity (-1.0 -- 1.0) and multiplied with 32767.0.
 * Otherwise the value is interpreted to be an integer and is assigned
 * as it is.
 */
void setConfigRangeInt (int * vp, ConfigContext * cfg) {

  assert (vp != NULL);
  assert (cfg != NULL);

  if (strchr (cfg->value, '.') != NULL) {
    double d;
    if (sscanf (cfg->value, "%lf", &d) == 1) {
      *vp = (int) (32767.0 * d);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else {
    int k;
    if (sscanf (cfg->value, "%d", &k) == 1) {
      *vp = k;
    }
    else {
      configIntUnparsable (cfg);
    }
  }

}

/*
 * Assigns an integer from the configuration context's value.
 */
void setConfigInt (int * vp, ConfigContext * cfg) {
  int k;
  assert (vp != NULL);
  assert (cfg != NULL);
  if (sscanf (cfg->value, "%d", &k) == 1) {
    *vp = k;
  }
  else {
    configIntUnparsable (cfg);
  }
}

/*
 * Assigns a double from the configuration context's value.
 */
void setConfigDouble (double * vp, ConfigContext * cfg) {
  double d;
  assert (vp != NULL);
  assert (cfg != NULL);
  if (sscanf (cfg->value, "%lf", &d) == 1) {
    *vp = d;
  }
  else {
    configDoubleUnparsable (cfg);
  }
}

/*
 * Returns the configuration entry's value string.
 */
char * getConfigValue (ConfigContext * cfg) {
  return cfg->value;
}

/*
 * Helper function for other module's config routines.
 * This routine compares the supplied parameter name with the name
 * in the configuration context struct and when a match is found,
 * attempts to set the target of the supplied storage pointer.
 * @param par  Pointer to the parameter name string.
 * @param cfg  Pointer to the configuration context.
 * @param dp   Pointer the double to set if a match is found.
 * @returns    Zero if the parameter name did not match the name in
 *             the context, 1 if it did and an assignment was made
 *             -1 if the name matched but no assignment was made.
 */
int getConfigParameter_d (char * par, ConfigContext * cfg, double * dp) {

  assert (par != NULL);
  assert (cfg != NULL);
  assert (dp  != NULL);

  if (strcasecmp (cfg->name, par) == 0) {
    double a;
    if (sscanf (cfg->value, "%lf", &a) == 1) {
      *dp = a;
    }
    else {
      configDoubleUnparsable (cfg);
      return -1;
    }
    return 1;
  }
  return 0;
}

/*
 * Helper function for other module's config routines.
 * This routine compares the supplied parameter name with the name
 * in the configuration context struct and when a match is found,
 * attempts to set the target of the supplied storage pointer.
 * @param par  Pointer to the parameter name string.
 * @param cfg  Pointer to the configuration context.
 * @param ip   Pointer the int to set if a match is found.
 * @returns    Zero if the parameter name did not match the name in
 *             the context, 1 if it did and an assignment was made
 *             -1 if the name matched but no assignment was made.
 */
int getConfigParameter_i (char * par, ConfigContext * cfg, int * ip) {

  assert (par != NULL);
  assert (cfg != NULL);
  assert (ip  != NULL);

  if (strcasecmp (cfg->name, par) == 0) {
    int i;
    if (sscanf (cfg->value, "%d", &i) == 1) {
      *ip = i;
    }
    else {
      configIntUnparsable (cfg);
      return -1;
    }
    return 1;
  }
  return 0;
}

/*
 * Helper function for other module's config routines.
 * This routine compares the supplied parameter name with the name
 * in the configuration context struct and when a match is found,
 * attempts to set the target of the supplied storage pointer.
 * @param par  Pointer to the parameter name string.
 * @param cfg  Pointer to the configuration context.
 * @param fp   Pointer the float to set if a match is found.
 * @returns    Zero if the parameter name did not match the name in
 *             the context, 1 if it did and an assignment was made
 *             -1 if the name matched but no assignment was made.
 */
int getConfigParameter_f (char * par, ConfigContext * cfg, float * fp) {

  assert (par != NULL);
  assert (cfg != NULL);
  assert (fp  != NULL);

  if (strcasecmp (cfg->name, par) == 0) {
    float a;
    if (sscanf (cfg->value, "%f", &a) == 1) {
      *fp = a;
    }
    else {
      configDoubleUnparsable (cfg);
      return -1;
    }
    return 1;
  }
  return 0;
}

/*
 * Range-checking version of getConfigParameter_i(...)
 * @param par  Pointer to the parameter name string.
 * @param cfg  Pointer to the configuration context.
 * @param fp   Pointer the float to set if a match is found.
 * @param lowInc Low limit (inclusive)
 * @param highInc High limit (inclusive)
 * @returns    Zero if the parameter name did not match the name in
 *             the context, 1 if it did and an assignment was made
 *             -1 if the name matched but no assignment was made.
 */
int getConfigParameter_ir (char * par,
			   ConfigContext * cfg,
			   int * ip,
			   int lowInc,
			   int highInc) {
  int k;
  int rtn;

  assert (ip != NULL);
  assert (lowInc <= highInc);

  rtn = getConfigParameter_i (par, cfg, &k);
  if (rtn == 1) {
    if ((lowInc <= k) && (k <= highInc)) {
      *ip = k;
    }
    else {
      printf ("Value for config parameter %s is out range (%d--%d).",
	      cfg->name,
	      lowInc,
	      highInc);
      rtn = -1;
    }
  }
  return rtn;
}

/*
 * Double-type with range-checking.
 * @returns    Zero if the parameter name did not match the name in
 *             the context, 1 if it did and an assignment was made
 *             -1 if the name matched but no assignment was made.
 */
int getConfigParameter_dr (char * par,
			   ConfigContext * cfg,
			   double * dp,
			   double lowInc,
			   double highInc) {
  double d;
  int rtn;

  assert (dp != NULL);
  assert (lowInc <= highInc);

  rtn = getConfigParameter_d (par, cfg, &d);
  if (rtn == 1) {
    if ((lowInc <= d) && (d <= highInc)) {
      *dp = d;
    }
    else {
      configDoubleOutOfRange (cfg, lowInc, highInc);
      rtn = -1;
    }
  }
  return rtn;
}

/*
 * Float-type with range-checking.
 * @returns    Zero if the parameter name did not match the name in
 *             the context, 1 if it did and an assignment was made
 *             -1 if the name matched but no assignment was made.
 */
int getConfigParameter_fr (char * par,
			   ConfigContext * cfg,
			   float * fp,
			   float lowInc,
			   float highInc) {
  double d;
  int rtn;

  assert (fp != NULL);
  assert (lowInc <= highInc);

  rtn = getConfigParameter_dr (par, cfg, &d, (double) lowInc, (double)highInc);
  if (rtn == 1) {
    *fp = d;
  }
  return rtn;
}

#ifdef CFG_MAIN
int main () {
  parseConfigurationFile ("test.cfg");
  return 0;
}
#endif /* CFG_MAIN */
