#ifndef MIDICTRL_H
#define MIDICTRL_H

/*
 * The enum type below defines a symbol for each Beatrix function that can be
 * influenced by a MIDI controller.
 *
 * When fully deployed, each module that registers a MIDI controller
 * listener with the MIDI parser indicates one of these symbols to inform
 * the parser which function is being handled. That way the MIDI parser
 * can use its configuration mapping and bind the function to the
 * appropriate MIDI channel and controller number.
 *
 * The handler function should have the following signature:
 * void f (unsigned char)
 * It will receive the 7-bit (0-127) data byte of the controller message.
 * The listener function is executed by the MIDI parser thread, so it is
 * wise to keep the execution time of the listener short.
 *
 * 14-bit controllers are not used in Beatrix, and many MIDI controllers
 * do not bother to send them either. However, here is how to do it in
 * this framework:
 *
 * 14-bit controllers (0-31 MSB) and (32-63 LSB) must be implemented
 * using a static data variable and two listeners where one updates the most
 * significant half of the variable and the other the least significant half.
 * Acting upon a controller update should of course be the duty of another
 * function, available to both listeners. For example:
 *
 * short data;
 * void update (short sv) {
 *   float a = ((float) sv) / 16383.0;
 *   ...
 * }
 * void fmsb (unsigned char uc) {
 *   data = (uc << 7) | (data & 0x7F);
 *   update (data);
 * }
 * void flsb (unsigned char uc) {
 *   data = (data & 0x03F8) | uc;
 *   update (data);
 * }
 *
 * You may also find it interesting to note that Pitch Bend is not a 
 * controller, but a separate 14-bit MIDI message. Thus it is not mapped
 * by this mechanism.
 */

enum midiCtrlFunction {
  ctSwellPedal,
  ctRotarySpeedToggle,
  ctRotarySpeed,

  ctUpperDrawbar0,
  ctUpperDrawbar1,
  ctUpperDrawbar2,
  ctUpperDrawbar3,
  ctUpperDrawbar4,
  ctUpperDrawbar5,
  ctUpperDrawbar6,
  ctUpperDrawbar7,
  ctUpperDrawbar8,

  ctLowerDrawbar0,
  ctLowerDrawbar1,
  ctLowerDrawbar2,
  ctLowerDrawbar3,
  ctLowerDrawbar4,
  ctLowerDrawbar5,
  ctLowerDrawbar6,
  ctLowerDrawbar7,
  ctLowerDrawbar8,

  ctPedalDrawbar0,
  ctPedalDrawbar1,
  ctPedalDrawbar2,
  ctPedalDrawbar3,
  ctPedalDrawbar4,
  ctPedalDrawbar5,
  ctPedalDrawbar6,
  ctPedalDrawbar7,
  ctPedalDrawbar8,
};

#endif /* MIDICTRL_H */
