/* whirl.c
 * Based on the scanner code, another Leslie experiment based on multiple
 * writing pointers.
 *
 * To do, in no particular order:
 * +Drum
 * +Filters
 * +Resonance (for horn)
 * Resonance for drum compartment
 * Fine-tuning
 * +Speed switching
 * +External controls
 * +Configuration interface
 * +De-interpolation noise (sample width)
 *
 * 12-sep-2004/FK Inserted two all-pass filters at the composite output
 *                hoping to de-correlate the channels; inconclusive results.
 *                See ifdef EXIT_FILTER.
 *
 * 23-aug-2004/FK (Re)introduced MIDI control of horn filter A's frequency.
 *                I needed to hear the difference in realtime. There is a
 *                distinct resonance at 3900 Hz (current default) but some
 *                extra brightness at 6000 Hz. Since it can also be set from
 *                config I will have to continue testing in rehearsal.
 * 21-apr-2004/FK Interpolating with a sample width of 1 sounds crappy,
 *                mostly because the lower frequencies introduce noise.
 *                However, by placing the low-pass filter at the drum
 *                output, much of the noise is reduced to an acceptable
 *                level.
 * 17-apr-2004/FK Sounds like a chorus.
 * 15-apr-2004/FK First version
 * 14-apr-2004/FK Opened file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <sys/soundcard.h>
#include <assert.h>

#include "main.h"
#include "config.h"
#include "eqcomp.h"
#include "radium.h"
#include "midi.h"
#include "whirl.h"

#ifdef WHIRL_MAIN
double SampleRateD = 22050.0;
#endif /* WHIRL_MAIN */

#define DISPLC_SIZE ((unsigned int) (1 << 11))
/* The DISPLC_MASK is a fixed point mask */
#define DISPLC_MASK ((DISPLC_SIZE << 16) - 1)

#define BUF_SIZE_SAMPLES ((unsigned int) (1 << 11))
#define BUF_MASK_SAMPLES (BUF_SIZE_SAMPLES - 1)


/* 
 * Forward (clockwise) displacement table for writing positions.
 */

static float hnFwdDispl[DISPLC_SIZE]; /* Horn */
static float drFwdDispl[DISPLC_SIZE]; /* Drum */

/*
 * Backward (counter-clockwise) displacement table. 
 */

static float hnBwdDispl[DISPLC_SIZE]; /* Horn */
static float drBwdDispl[DISPLC_SIZE]; /* Drum */

/* static float afw0[DISPLC_SIZE];	/\* Angle-dependent filtering *\/ */
/* static float afw1[DISPLC_SIZE];	/\* Forward *\/ */

/* static float abw0[DISPLC_SIZE];	/\* Angle-dependent filtering *\/ */
/* static float abw1[DISPLC_SIZE]; */

static float bfw[DISPLC_SIZE][5];
static float bbw[DISPLC_SIZE][5];

#define AGBUF 512
#define AGMASK (AGBUF-1)

static float adx0[AGBUF];
static float adx1[AGBUF];
static float adx2[AGBUF];
static int   adi0;
static int   adi1;
static int   adi2;

static double ipx;
static double ipy;


/*
 * Writing positions (actually, indexes into hnFwdDispl[]):
 *                Left  Right
 * Primary           0      1
 * First reflec.     2      3
 * Second refl.      4      5
 */

static int hornPhase[6];

static int drumPhase[6];

/* The current angle of rotating elements */

static int hornAngle = 0;
static int drumAngle = 0;

/* Added to angle. */
#define HI_SLOW 0.08
#define HI_FAST 0.58
static float hornIncrement = HI_SLOW;
#define LO_SLOW 0.06
#define LO_FAST 0.40
static float drumIncrement = LO_SLOW;

static float hornRPMslow = 40.0;
static float hornRPMfast = 400.0;
static float drumRPMslow = 30.0;
static float drumRPMfast = 330.0;

static float hornAcc = 2.0;	/* Revs per sec */
static float hornDec = 2.6;
static float drumAcc = 0.7;
static float drumDec = 0.6;

typedef struct _revcontrol {
  int hornTarget;
  int drumTarget;
} RevControl;

static RevControl revoptions [9];
static int revselects[9];
static size_t revSelectEnd;
static int revSelect = 0;

static int hornAcDc = 0;
static int drumAcDc = 0;

static int hornIncrUI = 0;
static int drumIncrUI = 0;

static int hornAII = 0;
static int drumAII = 0;

static int hornTarget = 0;
static int drumTarget = 0;

static int hornAIIAcc;
static int hornAIIDec;
static int drumAIIAcc;
static int drumAIIDec;

/*
 * Spacing between reflections in samples. The first can't be zero, since
 * we must allow for the swing of the extent to wander close to the reader.
 */

static float hornSpacing[6] = {
  12.0,				/* Primary */
  18.0,
  53.0,				/* First reflection */
  50.0,
  106.0,			/* Secondary reflection */
  116.0
};

static float hornRadiusCm = 19.2; /* 17.0; 25-nov-04 */
static float drumRadiusCm = 22.0;

static float airSpeed = 340.0;	/* Meters per second */

static float micDistCm = 42.0;	/* From mic to origin */

static float cabinetWidthCm = 74.0;
static float cabinetDepthCm = 52.0;

static float drumSpacing[6] = {
  36.0, 39.0,
  79.0, 86.0,
  123.0, 116.0
};

/* Delay buffers */

static float HLbuf[BUF_SIZE_SAMPLES]; /* Horn left buffer */
static float HRbuf[BUF_SIZE_SAMPLES]; /* Horn right buffer */
static float DLbuf[BUF_SIZE_SAMPLES]; /* Drum left buffer */
static float DRbuf[BUF_SIZE_SAMPLES]; /* Drum right buffer */

/* Single read position, incremented by one, always. */

static unsigned int outpos = 0;

typedef enum {a0, a1, a2, b0, b1, b2, z0, z1} filterCoeff;

static float drfw[8];		/* Drum filter */
static float drfw2[8];		/* Drum filter for post */
static float * drfL = drfw;	/* Left drum filter array */
static float * drfR = drfw2;	/* Right drum filter array */
static int    lpT = 8;
static double lpF = 811.9695;	/* Frequency */
static double lpQ =   1.6016;	/* Q, bandwidth */
static double lpG = -38.9291;	/* Gain */

static float hafw[8];		/* Horn filter a */
static float haT = 0;		/* low pass */
static float haF = 4500;  /* 3900.0; 25-nov-04 */
static float haQ = 2.7456; /*   1.4; 25-nov-04 */
static float haG = -30.0;  /*   0.0; 25-nov-04 */

static float hbfw[8];
static float hbT = 7;		/* low shelf */
static float hbF = 300.0;
static float hbQ =   1.0; /* 2.0; 25-nov-04 */
static float hbG = -30.0; /* -60.0; 25-nov-04 */ /* negative gain */

#define COMB_SIZE ((unsigned int) (1 << 10))
#define COMB_MASK (COMB_SIZE - 1)

static float comb0[COMB_SIZE];
static float cb0fb = -0.55;
static int   cb0dl = 38;
static float * cb0wp;		/* Write pointer */
static float * cb0rp;		/* Read pointer */
static float * cb0bp;		/* Begin pointer */
static float * cb0es;		/* End sentinel */

static float comb1[COMB_SIZE];
static float cb1fb = -0.3508;
static int   cb1dl = 120;
static float * cb1wp;
static float * cb1rp;
static float * cb1bp;
static float * cb1es;

static float hornLevel = 0.7;
static float leakLevel = 0.15;
static float leakage = 0;
static float outputLevel = 0.5;	/* Not used */

/* #define EXIT_FILTER */
#ifdef EXIT_FILTER
static float xlfw[8];		/* Exit filter left channel */
static float xlp[4] = {EQC_APF, 1000.0, 1.0, 0.0};
static float xrfw[8];
static float xrp[4] = {EQC_APF, 1000.0, 1.0, 0.0};
#endif /* EXIT_FILTER */

/*
 *
 */
static int fixpt (float u) {
  int f;
  int m;
  f = (int) floorf (u);
  m = (int) (65536.0 * (u - floorf (u)));
  return (f << 16) | m;
}

static int fixptd (double u) {
  int f;
  int m;
  f = (int) floor (u);
  m = (int) (65536.0 * (u - floor (u)));
  return (f << 16) | m;
}

/*
 *
 */
static void setIIRFilter (float W[],
			  int T,
			  const double F,
			  const double Q,
			  const double G) {
  double C[6];
  eqCompute (T, F, Q, G, C);
  W[a1] = C[EQC_A1];
  W[a2] = C[EQC_A2];
  W[b0] = C[EQC_B0];
  W[b1] = C[EQC_B1];
  W[b2] = C[EQC_B2];
}

/*
 * Sets the revolution selection.
 */
void setRevSelect (int n) {
  int i;

  revSelect = n % revSelectEnd;
  i = revselects[revSelect];

#if 0
  printf ("\rREV:%d %d ", n, i); fflush (stdout);
#endif

  hornTarget = revoptions[i].hornTarget;
  drumTarget = revoptions[i].drumTarget;

  if (hornIncrUI < hornTarget) {
    hornAII = hornAIIAcc;
    hornAcDc = 1;
  }
  else if (hornTarget < hornIncrUI) {
    hornAII = hornAIIDec;
    hornAcDc = -1;
  }
  if (drumIncrUI < drumTarget) {
    drumAII = drumAIIAcc;
    drumAcDc = 1;
  }
  else if (drumTarget < drumIncrUI) {
    drumAII = drumAIIDec;
    drumAcDc = -1;
  }
  
}

void advanceRevSelect () {
  setRevSelect (revSelect + 1);
}

static void revControl (unsigned char u) {
  setRevSelect ((int) (u / 16));
}

/* 11-may-2004/FK Hack to make the pedal work. See midiIn.c */
void setWhirlSustainPedal (unsigned char u) {
  if (u) {
    setRevSelect ((revSelect == 1) ? 2 : 1);
  }
}

/*
 * Returns the increment needed to traverse a certain table the
 * requested number of revolutions per minute.
 * @param tblSz  The number of slots in the table.
 * @param rpm    Revolutions per minute.
 */
static double getTableInc (size_t tblSz, double rpm) {
  return (rpm * (double) tblSz) / (SampleRateD * 60.0);
}

static int ui_increment (double rpm) {
  return fixptd (getTableInc ((size_t) DISPLC_SIZE, rpm));
}

static int incrementChange (float chg) {
  double fragsz = 128.0;
  double a = fragsz / (chg * SampleRateD);
  return ui_increment (a);
}

static void setRevOption (int i,
			  unsigned int hnTgt,
			  unsigned int drTgt) {
  revoptions[i].hornTarget = hnTgt;
  revoptions[i].drumTarget = drTgt;
}

static void computeRotationSpeeds () {
  int hfast = ui_increment (hornRPMfast);
  int hslow = ui_increment (hornRPMslow);
  int hstop = 0;
  int dfast = ui_increment (drumRPMfast);
  int dslow = ui_increment (drumRPMslow);
  int dstop = 0;

  hornAIIAcc = ui_increment (hornAcc);
  hornAIIDec = ui_increment (hornDec);
  drumAIIAcc = ui_increment (drumAcc);
  drumAIIDec = ui_increment (drumDec);

  setRevOption (8, hfast, dfast);
  setRevOption (7, hfast, dslow);
  setRevOption (6, hfast, dstop);
  setRevOption (5, hslow, dfast);
  setRevOption (4, hslow, dslow);
  setRevOption (3, hslow, dstop);
  setRevOption (2, hstop, dfast);
  setRevOption (1, hstop, dslow);
  setRevOption (0, hstop, dstop);

  revselects[0] = 0;		/* stop */
  revselects[1] = 4;		/* both slow */
  revselects[2] = 8;		/* both fast */
  revselects[3] = 4;		/* both slow */
  revSelectEnd = 4;		/* How may entries in revselects[] */
  revSelect = 1;		/* Where we start */
}

/*
 *
 */
static void ipolmark (double degrees, double level) {
  ipx = degrees;
  ipy = level;
}

/*
 *
 */
static void ipoldraw (double degrees, double level, int partial) {
  double d;
  double e;
  double apos;
  double range;
  int fromIndex;
  int toIndex;
  int i;

  d = ipx;
  while (d < 0.0) d += 360.0;
  fromIndex = (int) ((d * (double) DISPLC_SIZE) / 360.0);

  ipx = degrees;

  e = ipx;
  while (e < d) e += 360.0;
  toIndex = (int) ((e * (double) DISPLC_SIZE) / 360.0);

  range = (double) (toIndex - fromIndex);
  for (i = fromIndex; i <= toIndex; i++) {
    double x = (double) (i - fromIndex);
    double w = ipy + ((x / range) * (level - ipy));
    bfw[i & (DISPLC_SIZE - 1)][partial] = (float) w;
  }
  
  ipy = level;
} /* ipoldraw */

/*
 *
 */
static void initTables () {
  int i;
  int j;
  double sum;
  double hornRadiusSamples = (hornRadiusCm * 220.5) / airSpeed;
  double drumRadiusSamples = (drumRadiusCm * 220.5) / airSpeed;
  double micDistSamples    = (micDistCm    * 220.5) / airSpeed;

  for (i = 0; i < DISPLC_SIZE; i++) {
    /* Compute angle around the circle */
    double v = (2.0 * M_PI * (double) i) / (double) DISPLC_SIZE;
    /* Distance between the mic and the rotor korda */
    double a = micDistSamples - (hornRadiusSamples * cos (v));
    /* Distance between rotor and mic-origin line */
    double b = hornRadiusSamples * sin (v);

    hnFwdDispl[i] = sqrt ((a * a) + (b * b));
    hnBwdDispl[DISPLC_SIZE - (i + 1)] = hnFwdDispl[i];

    a = micDistSamples - (drumRadiusSamples * cos (v));
    b = drumRadiusSamples * sin (v);
    drFwdDispl[i] = sqrt ((a * a) + (b * b));
    drBwdDispl[DISPLC_SIZE - (i + 1)] = drFwdDispl[i];

    a = (1.0 + cos (v)) / 2.0;	/* Goes from 1, 0.5, 0, 0.5, 1 */
#ifdef COMMENT
    afw0[i] = 0.5 * (1.0 + a);
    afw1[i] = 1.0 - afw0[i];

#if 1
    afw0[i] *= 0.33321 + (0.33321 * a);
    afw1[i] *= 0.33321 * (1.0 - a);
#endif

    abw0[DISPLC_SIZE - (i + 1)] = afw0[i];
    abw1[DISPLC_SIZE - (i + 1)] = afw1[i];
#endif /* COMMENT */
  }

  hornPhase[0] = 0;
  hornPhase[1] = DISPLC_MASK >> 1;

  hornPhase[2] = ((DISPLC_SIZE * 2) / 6) << 16;
  hornPhase[3] = ((DISPLC_SIZE * 5) / 6) << 16;

  hornPhase[4] = ((DISPLC_SIZE * 1) / 6) << 16;
  hornPhase[5] = ((DISPLC_SIZE * 4) / 6) << 16;

  for (i = 0; i < 6; i++) {
    hornSpacing[i] += hornRadiusSamples + 1.0;
  }

  drumPhase[0] = 0;
  drumPhase[1] = DISPLC_MASK >> 1;

  drumPhase[2] = ((DISPLC_SIZE * 2) / 6) << 16;
  drumPhase[3] = ((DISPLC_SIZE * 5) / 6) << 16;

  drumPhase[4] = ((DISPLC_SIZE * 1) / 6) << 16;
  drumPhase[5] = ((DISPLC_SIZE * 4) / 6) << 16;

  for (i = 0; i < 6; i++) {
    drumSpacing[i] += drumRadiusSamples + 1.0;
  }

  setIIRFilter (drfw, lpT, lpF, lpQ, lpG);
  setIIRFilter (drfw2, lpT, lpF, lpQ, lpG);
  setIIRFilter (hafw, haT, haF, haQ, haG);
  setIIRFilter (hbfw, hbT, hbF, hbQ, hbG);

#ifdef EXIT_FILTER
  setIIRFilter (xlfw, xlp[0], xlp[1], xlp[2], xlp[3]);
  setIIRFilter (xrfw, xrp[0], xrp[1], xrp[2], xrp[3]);
#endif /* EXIT_FILTER */

  cb0rp = &(comb0[COMB_SIZE - cb0dl]);
  cb0wp = &(comb0[0]);
  cb0bp = &(comb0[0]);
  cb0es = &(comb0[COMB_SIZE]);

  cb1rp = &(comb1[COMB_SIZE - cb1dl]);
  cb1wp = &(comb1[0]);
  cb1bp = &(comb1[0]);
  cb1es = &(comb1[COMB_SIZE]);

  computeRotationSpeeds ();

  /* Horn angle-dependent impulse response coefficents. */
  /* These were derived from 'Doppler simulation and the leslie',
   * Julius Smith, Stefania Serafin, Jonathan Abel, David Berners,
   * Proc. of the 5th Conference on Digital Audio Effects (DAFx-02),
   * Hamburg, Germany, September 26-28, 2002.
   * In this article figure 8 depicts the 'First 5 principal components
   * weighted by their corresponding singular values'. I have been unable
   * to clearly understand what this means, but anyway ploughed ahead
   * interpreted the components as impulse response coefficients. The
   * figure plots the five components as horizontal wavy lines, where
   * the x axis is the angle and the y axis is relative to each component.
   * So, the following code consists of 'drawing' instructions that simply
   * 'paints' a copy of each line as interpreted from the figure, from
   * left (-180 degrees) to right (180 degrees). The points between the
   * given coordinates are linearly interpolated and inserted into the
   * bfw matrix. Then (below) we normalise the lot to unit gain.
   */
  ipolmark (-180.0, 1.10);
  ipoldraw (-166.3, 0.85, 0);
  ipoldraw (-150.0, 0.85, 0);
  ipoldraw (-134.8, 0.85, 0);
  ipoldraw (-123.6, 0.78, 0);
  ipoldraw (-105.6, 0.78, 0);
  ipoldraw (-100.0, 0.78, 0);
  ipoldraw ( -78.6, 0.85, 0);
  ipoldraw ( -60.7, 0.81, 0);
  ipoldraw ( -50.0, 0.89, 0);
  ipoldraw ( -44.7, 0.90, 0);
  ipoldraw ( -29.2, 1.30, 0);
  ipoldraw (   0.0, 2.79, 0);
  ipoldraw (  15.7, 2.30, 0);
  ipoldraw (  31.5, 1.50, 0);
  ipoldraw (  44.9, 0.90, 0);
  ipoldraw (  50.0, 0.88, 0);
  ipoldraw (  60.7, 0.80, 0);
  ipoldraw (  74.2, 0.84, 0);
  ipoldraw ( 100.0, 0.75, 0);
  ipoldraw ( 121.3, 0.80, 0);
  ipoldraw ( 150.0, 0.87, 0);
  ipoldraw ( 166.3, 0.83, 0);
  ipoldraw ( 180.0, 1.10, 0);

  ipolmark (-180.0, -0.10);
  ipoldraw (-150.0,  0.10, 1);
  ipoldraw (-137.0, -0.10, 1);
  ipoldraw (-118.0,  0.18, 1);
  ipoldraw (-100.0,  0.20, 1);
  ipoldraw ( -87.0,  0.40, 1);
  ipoldraw ( -75.0,  0.30, 1);
  ipoldraw ( -44.5,  0.70, 1);
  ipoldraw ( -30.0,  0.45, 1);
  ipoldraw (   3.1, -0.75, 1);
  ipoldraw (  15.6, -0.51, 1);
  ipoldraw (  35.7,  0.45, 1);
  ipoldraw (  44.6,  0.67, 1);
  ipoldraw (  74.7,  0.20, 1);
  ipoldraw (  90.2,  0.31, 1);
  ipoldraw ( 107.0,  0.02, 1);
  ipoldraw ( 122.3,  0.15, 1);
  ipoldraw ( 135.3, -0.12, 1);
  ipoldraw ( 153.3,  0.08, 1);
  ipoldraw ( 180.0, -0.10, 1);

  ipolmark (-180.0,  0.40);
  ipoldraw (-165.0,  0.20, 2);
  ipoldraw (-150.0,  0.48, 2);
  ipoldraw (-121.2,  0.22, 2);
  ipoldraw ( -89.2,  0.30, 2);
  ipoldraw ( -69.2,  0.22, 2);
  ipoldraw ( -58.0,  0.11, 2);
  ipoldraw ( -40.2, -0.43, 2);
  ipoldraw ( -29.0, -0.58, 2);
  ipoldraw ( -15.6, -0.49, 2);
  ipoldraw (   0.0,  0.00, 2);
  ipoldraw (  17.8, -0.60, 2);
  ipoldraw (  75.9,  0.35, 2);
  ipoldraw (  91.5,  0.28, 2);
  ipoldraw ( 104.9,  0.32, 2);
  ipoldraw ( 122.7,  0.22, 2);
  ipoldraw ( 150.0,  0.45, 2);
  ipoldraw ( 167.0,  0.20, 2);
  ipoldraw ( 180.0,  0.40, 2);

  ipolmark (-180.0, -0.15);
  ipoldraw (-165.2, -0.20, 3);
  ipoldraw (-150.0,  0.00, 3);
  ipoldraw (-133.9, -0.20, 3);
  ipoldraw (-106.0,  0.09, 3);
  ipoldraw ( -89.3, -0.20, 3);
  ipoldraw ( -76.3,  0.00, 3);
  ipoldraw ( -60.3,  0.34, 3);
  ipoldraw ( -44.6,  0.00, 3);
  ipoldraw ( -15.6, -0.28, 3);
  ipoldraw (   0.0,  0.36, 3);
  ipoldraw (  14.5,  0.15, 3);
  ipoldraw (  20.1, -0.18, 3);
  ipoldraw (  44.6,  0.18, 3);
  ipoldraw (  60.4,  0.26, 3);
  ipoldraw (  75.9,  0.20, 3);
  ipoldraw (  90.4, -0.10, 3);
  ipoldraw ( 104.9,  0.10, 3);
  ipoldraw ( 122.8, -0.10, 3);
  ipoldraw ( 136.2, -0.10, 3);
  ipoldraw ( 150.0,  0.12, 3);
  ipoldraw ( 165.0, -0.25, 3);
  ipoldraw ( 180.0, -0.15, 3);

  ipolmark (-180.0,  0.20);
  ipoldraw (-165.2,  0.00, 4);
  ipoldraw (-150.0,  0.22, 4);
  ipoldraw (-136.2, -0.22, 4);
  ipoldraw (-120.5,  0.00, 4);
  ipoldraw (-100.0,  0.00, 4);
  ipoldraw ( -90.0,  0.10, 4);
  ipoldraw ( -74.8, -0.15, 4);
  ipoldraw ( -60.3, -0.20, 4);
  ipoldraw ( -44.6,  0.21, 4);
  ipoldraw ( -15.6,  0.00, 4);
  ipoldraw (   0.0,  0.30, 4);
  ipoldraw (  15.6, -0.21, 4);
  ipoldraw (  20.1, -0.11, 4);
  ipoldraw (  45.7,  0.18, 4);
  ipoldraw (  60.3, -0.10, 4);
  ipoldraw (  74.8, -0.20, 4);
  ipoldraw (  90.4, -0.05, 4);
  ipoldraw ( 104.9, -0.18, 4);
  ipoldraw ( 120.5,  0.00, 4);
  ipoldraw ( 136.2, -0.30, 4);
  ipoldraw ( 150.0,  0.20, 4);
  ipoldraw ( 165.0,  0.00, 4);
  ipoldraw ( 180.0,  0.20, 4);

  sum = 0.0;
  /* Compute the normalisation factor */
  for (i = 0; i < DISPLC_SIZE; i++) {
    double colsum = 0.0;
    for (j = 0; j < 5; j++) {
      colsum += fabs (bfw[i][j]);
    }
    if (sum < colsum) {
      sum = colsum;
    }
  }
  /* Apply normalisation */
  for (i = 0; i < DISPLC_SIZE; i++) {
    for (j = 0; j < 5; j++) {
      bfw[i][j] *= 1.0 / sum;
      bbw[DISPLC_SIZE - i - 1][j] = bfw[i][j];
    }
  }
  
} /* initTables */

/*
 * Displays the settings of a filter.
 */
static void displayFilter (char * id, int T, float F, float Q, float G) {
  char * type = eqGetTypeString (T);
  printf ("\r%s:T=%3.3s:F=%10.4f:Q=%10.4f:G=%10.4f", id, type, F, Q, G);
  fflush (stdout);
}

#define UPDATE_A_FILTER { \
  setIIRFilter (hafw, haT, haF, haQ, haG); \
  displayFilter ("Horn A", haT, haF, haQ, haG); \
}


/*
 * Sets the type of the A horn IIR filter.
 */
static void setHornFilterAType (unsigned char uc) {
  haT = (int) (uc / 15);
  UPDATE_A_FILTER;
}

/*
 * Sets the cutoff frequency of the A horn IIR filter.
 */
static void setHornFilterAFrequency (unsigned char uc) {
  double u = (double) uc;
  double minv = 250.0;
  double maxv = 8000.0;
  haF = minv + ((maxv - minv) * ((u * u) / 16129.0));
  UPDATE_A_FILTER;
}

/*
 * Sets the Q value of the A horn IIR filter.
 */
static void setHornFilterAQ (unsigned char uc) {
  double u = (double) uc;
  double minv = 0.01;
  double maxv = 6.00;
  haQ = minv + ((maxv - minv) * (u / 127.0));
  UPDATE_A_FILTER;
}

/*
 * Sets the Gain value of the A horn IIR filter.
 */
static void setHornFilterAGain (unsigned char uc) {
  double u = (double) uc;
  double minv = -48.0;
  double maxv =  48.0;
  haG = minv + ((maxv - minv) * (u / 127.0));
  UPDATE_A_FILTER;
}

#ifdef EXIT_FILTER
static double getFilterFrequency (double u, double minv, double maxv) {
  return minv + ((maxv - minv) * ((u * u) / 16129.0));
}
static void setIIRFilterEx (float W[], float P[]) {
  setIIRFilter (W, P[0], P[1], P[2], P[3]);
}
static void setExitFilterLeftFrequency (unsigned char uc) {
  xlp[1] = (float) getFilterFrequency ((double) uc, 250.0, 6000.0);
  setIIRFilterEx (xlfw, xlp);
}
static void setExitFilterRightFrequency (unsigned char uc) {
  xrp[1] = (float) getFilterFrequency ((double) uc, 250.0, 6000.0);
  setIIRFilterEx (xrfw, xrp);
}
#endif /* EXIT_FILTER */

/*
 * This function initialises this module. It is run after whirlConfig.
 */
void initWhirl () {

  initTables ();

  hornIncrUI = fixpt (hornIncrement);
  drumIncrUI = fixpt (drumIncrement);

  leakage = leakLevel * hornLevel;

#ifndef WHIRL_MAIN
#if 1

  useMIDIControlFunction ("rotary.speed-toggle",    setWhirlSustainPedal);
  useMIDIControlFunction ("rotary.speed-select",    revControl);

  useMIDIControlFunction ("whirl.horn.filter.a.type", setHornFilterAType);
  useMIDIControlFunction ("whirl.horn.filter.a.hz",   setHornFilterAFrequency);
  useMIDIControlFunction ("whirl.horn.filter.a.q",    setHornFilterAQ);
  useMIDIControlFunction ("whirl.horn.filter.a.gain", setHornFilterAGain);

#else

  /* CTL_SUSTAIN is a symbol from sys/soundcard.h */
  setMIDIControllerFunction (CTL_SUSTAIN, setWhirlSustainPedal);
  /* Symbols from radium.h */
  setMIDIControllerFunction (RD_P1_K16, revControl);

  setMIDIControllerFunction (RD_P3_S02, setHornFilterAFrequency);
  setMIDIControllerFunction (RD_P3_S03, setHornFilterAQ);

#endif

#ifdef EXIT_FILTER
  setMIDIControllerFunction (RD_P3_K09, setExitFilterLeftFrequency);
  setMIDIControllerFunction (RD_P3_K10, setExitFilterRightFrequency);
#endif /* EXIT_FILTER */
#endif /* no WHIRL_MAIN */
}

/*
 * Configuration interface.
 */
int whirlConfig (ConfigContext * cfg) {
  double d;
  int k;
  int rtn = 1;
  if (getConfigParameter_d ("whirl.horn.slowrpm", cfg, &d) == 1) {
    hornRPMslow = (float) d;
  }
  else if (getConfigParameter_d ("whirl.horn.fastrpm", cfg, &d) == 1) {
    hornRPMfast = (float) d;
  }
  else if (getConfigParameter_d ("whirl.horn.accelleration", cfg, &d) == 1) {
    hornAcc = (float) d;
  }
  else if (getConfigParameter_d ("whirl.horn.decelleration", cfg, &d) == 1) {
    hornDec = (float) d;
  }
  else if (getConfigParameter_d ("whirl.drum.slowrpm", cfg, &d) == 1) {
    drumRPMslow = (float) d;
  }
  else if (getConfigParameter_d ("whirl.drum.fastrpm", cfg, &d) == 1) {
    drumRPMfast = (float) d;
  }
  else if (getConfigParameter_d ("whirl.drum.accelleration", cfg, &d) == 1) {
    drumAcc = (float) d;
  }
  else if (getConfigParameter_d ("whirl.drum.decelleration", cfg, &d) == 1) {
    drumDec = (float) d;
  }
  else if (getConfigParameter_d ("whirl.horn.radius", cfg, &d) == 1) {
    hornRadiusCm = (float) d;
  }
  else if (getConfigParameter_d ("whirl.drum.radius", cfg, &d) == 1) {
    drumRadiusCm = (float) d;
  }
  else if (getConfigParameter_d ("whirl.horn.level", cfg, &d) == 1) {
    hornLevel = (float) d;
  }
  else if (getConfigParameter_d ("whirl.horn.leak", cfg, &d) == 1) {
    leakLevel = (float) d;
  }
  else if (getConfigParameter_ir
	   ("whirl.drum.filter.type", cfg, &k, 0, 8) == 1) {
    lpT = k;
  }
  else if (getConfigParameter_d ("whirl.drum.filter.q", cfg, &d) == 1) {
    lpQ =  d;
  }
  else if (getConfigParameter_d ("whirl.drum.filter.hz", cfg, &d) == 1) {
    lpF =  d;
  }
  else if (getConfigParameter_d ("whirl.drum.filter.gain", cfg, &d) == 1) {
    lpG =  d;
  }
  else if (getConfigParameter_ir
	   ("whirl.horn.filter.a.type", cfg, &k, 0, 8) == 1) {
    haT = k;
  }
  else if (getConfigParameter_d ("whirl.horn.filter.a.hz", cfg, &d) == 1) {
    haF = (double) d;
  }
  else if (getConfigParameter_d ("whirl.horn.filter.a.q", cfg, &d) == 1) {
    haQ = (double) d;
  }
  else if (getConfigParameter_d ("whirl.horn.filter.a.gain", cfg, &d) == 1) {
    haG = (double) d;
  }
  else if (getConfigParameter_ir
	   ("whirl.horn.filter.b.type", cfg, &k, 0, 8) == 1) {
    hbT = k;
  }
  else if (getConfigParameter_d ("whirl.horn.filter.b.hz", cfg, &d) == 1) {
    hbF = (double) d;
  }
  else if (getConfigParameter_d ("whirl.horn.filter.b.q", cfg, &d) == 1) {
    hbQ = (double) d;
  }
  else if (getConfigParameter_d ("whirl.horn.filter.b.gain", cfg, &d) == 1) {
    hbG = (double) d;
  }
  else if (getConfigParameter_d ("whirl.horn.comb.a.feedback", cfg, &d) == 1) {
    cb0fb = (double) d;
  }
  else if (getConfigParameter_i ("whirl.horn.comb.a.delay", cfg, &k) == 1) {
    cb0dl = k;
  }
  else if (getConfigParameter_d ("whirl.horn.comb.b.feedback", cfg, &d) == 1) {
    cb1fb = (double) d;
  }
  else if (getConfigParameter_i ("whirl.horn.comb.b.delay", cfg, &k) == 1) {
    cb1dl = k;
  }

  else {
    rtn = 0;
  }
  return rtn;
}

/*
 * 
 */
unsigned char * whirlProc (unsigned char * inbuffer,
			   unsigned char * outbuffer,
			   size_t bufferLengthSamples) {

  short * xp = (short *) inbuffer;
  short * yp = (short *) outbuffer;
  int i;
  static float z[12];

  if (hornAcDc) {
    if (0 < hornAcDc) {
      hornIncrUI += hornAII;
      if (hornTarget <= hornIncrUI) {
	hornIncrUI = hornTarget;
	hornAcDc = 0;
      }
    }
    else {
      hornIncrUI -= hornAII;
      if (hornIncrUI <= hornTarget) {
	hornIncrUI = hornTarget;
	hornAcDc = 0;
      }
    }
  }

  if (drumAcDc) {
    if (0 < drumAcDc) {
      drumIncrUI += drumAII;
      if (drumTarget <= drumIncrUI) {
	drumIncrUI = drumTarget;
	drumAcDc = 0;
      }
    }
    else {
      drumIncrUI -= drumAII;
      if (drumIncrUI <= drumTarget) {
	drumIncrUI = drumTarget;
	drumAcDc = 0;
      }
    }
  }

  for (i = 0; i < bufferLengthSamples; i++) {
    unsigned int k;
    unsigned int n;
    float q;
    float r;
    float w;
    float x = (float) (*xp++);
    float xa;
    float xx = x;
    float leak;

#define ANGFILTER(BW,DX,DI) {           \
    xa  = BW[k][0] * x;                     \
    xa += BW[k][1] * DX[(DI)];              \
    xa += BW[k][2] * DX[((DI)+1) & AGMASK]; \
    xa += BW[k][3] * DX[((DI)+2) & AGMASK]; \
    xa += BW[k][4] * DX[((DI)+3) & AGMASK]; \
    }

#define ADDHIST(DX,DI,XS) {      \
    DI = (DI + AGMASK) & AGMASK; \
    DX[DI] = XS;}

#if 1
    /* Assumes that input samples have width=1.0. Results in artefacts. */
#define HN_MOTION(P,BUF,DSP,BW,DX,DI) {                 \
    k = ((hornAngle + hornPhase[(P)]) & DISPLC_MASK) >> 16; \
    w = hornSpacing[(P)] + DSP[k] + (float) outpos;         \
    r = floorf (w);                                         \
    ANGFILTER(BW,DX,DI);                                \
    q = xa * (w - r);                                       \
    n = ((unsigned int) r) & BUF_MASK_SAMPLES;              \
    BUF[n] += xa - q;                                       \
    n = (n + 1) & BUF_MASK_SAMPLES;                         \
    BUF[n] += q;}

#define DR_MOTION(P,BUF,DSP) {                              \
    k = ((drumAngle + drumPhase[(P)]) & DISPLC_MASK) >> 16; \
    w = drumSpacing[(P)] + DSP[k] + (float) outpos;         \
    r = floorf (w);                                         \
    q = x * (w - r);                                        \
    n = ((unsigned int) r) & BUF_MASK_SAMPLES;              \
    BUF[n] += x - q;                                        \
    n = (n + 1) & BUF_MASK_SAMPLES;                         \
    BUF[n] += q;}

#else
    /*
     * Strangely enough, all my attempts to alleviate the under- and
     * overlap noise sounded worse than the above.
     */
#endif

    /* This is just a bum filter to take some high-end off. */
#if 1
#define FILTER_C(W0,W1,I) {           \
    float temp = x;                   \
    x = ((W0) * x) + ((W1) * z[(I)]); \
    z[(I)] = temp; }
#else
#define FILTER_C(W0,W1,I) { \
    x = ((W0) * x) + ((W1) * z[(I)]); \
    z[(I)] = x;}
#endif

#define EQ_IIR(W,X,Y) { \
    float temp = (X) - (W[a1] * W[z0]) - (W[a2] * W[z1]); \
    Y = (temp * W[b0]) + (W[b1] * W[z0]) + (W[b2] * W[z1]); \
    W[z1] = W[z0]; \
    W[z0] = temp;}

#define COMB(WP,RP,BP,ES,FB,X) {\
    X += ((*(RP)++) * (FB)); \
    *(WP)++ = X; \
    if ((RP) == (ES)) RP = BP; \
    if ((WP) == (ES)) WP = BP;}

    EQ_IIR(hafw, x, x);
    EQ_IIR(hbfw, x, x);

    leak = x * leakage;

    COMB(cb0wp, cb0rp, cb0bp,cb0es, cb0fb, x);
    COMB(cb1wp, cb1rp, cb1bp,cb1es, cb1fb, x);

    /* --- STATIC HORN FILTER --- */
#if 1
    /* HORN PRIMARY */
    HN_MOTION(0, HLbuf, hnFwdDispl, bfw, adx0, adi0);
    HN_MOTION(1, HRbuf, hnBwdDispl, bbw, adx0, adi0);
    ADDHIST(adx0, adi0, x);
#endif

#if 1
    /* --- HORN FIRST REFLECTION FILTER --- */
    FILTER_C(0.4, 0.4, 0);
/*     FILTER_C(0.4, 0.4, 0); */

    /* HORN FIRST REFLECTION */
    HN_MOTION(2, HLbuf, hnBwdDispl, bbw, adx1, adi1);
    HN_MOTION(3, HRbuf, hnFwdDispl, bfw, adx1, adi1);
    ADDHIST(adx1, adi1, x);
#endif
#if 1
    /* --- HORN SECOND REFLECTION FILTER --- */
    FILTER_C(0.4, 0.4, 1);
/*     FILTER_C(0.4, 0.4, 1); */

    /* HORN SECOND REFLECTION */
    HN_MOTION(4, HLbuf, hnFwdDispl, bfw, adx2, adi2);
    HN_MOTION(5, HRbuf, hnBwdDispl, bbw, adx2, adi2);
    ADDHIST(adx2, adi2, x);

#endif

    x = xx;
/*     EQ_IIR(drfw, x, x); */

#define DRUM
#ifdef DRUM
#if 1
    /* --- DRUM --- */
    DR_MOTION(0, DLbuf, drFwdDispl);
    DR_MOTION(1, DRbuf, drBwdDispl);
#endif
#if 1
    FILTER_C(0.4, 0.4, 2);
    DR_MOTION(2, DLbuf, drBwdDispl);
    DR_MOTION(3, DRbuf, drFwdDispl);
#endif
#if 1
    FILTER_C(0.4, 0.4, 3);
    DR_MOTION(4, DLbuf, drFwdDispl);
    DR_MOTION(5, DRbuf, drBwdDispl);
#endif
#endif /* DRUM */

#if 1
#ifdef EXIT_FILTER
    {
      float y;
      float yy;

      EQ_IIR(drfL, DLbuf[outpos], y);
      yy = (y + (hornLevel * HLbuf[outpos]) + leak);
      EQ_IIR(xlfw, yy, yy);
      *yp++ = yy;

      EQ_IIR(drfR, DRbuf[outpos], y);
      yy = (y + (hornLevel * HRbuf[outpos]) + leak);
      EQ_IIR(xrfw, yy, yy);
      *yp++ = yy;
    }
#else
    {
      float y;
      EQ_IIR(drfL, DLbuf[outpos], y);
      *yp++ = (short) (y + (hornLevel * HLbuf[outpos]) + leak);
      EQ_IIR(drfR, DRbuf[outpos], y);
      *yp++ = (short) (y + (hornLevel * HRbuf[outpos]) + leak);
    }
#endif /* EXIT_FILTER */
#else
    {
      float y;
      float mono;
      EQ_IIR(drfL, DLbuf[outpos], y);
      mono = (short) (y + (hornLevel * HLbuf[outpos]) + leak);
      EQ_IIR(drfR, DRbuf[outpos], y);
      mono += (short) (y + (hornLevel * HRbuf[outpos]) + leak);
      *yp++ = mono;
      *yp++ = mono;
    }
#endif

    HLbuf[outpos] = 0.0;
    HRbuf[outpos] = 0.0;
    DLbuf[outpos] = 0.0;
    DRbuf[outpos] = 0.0;

    outpos = (outpos + 1) & BUF_MASK_SAMPLES;
    hornAngle = (hornAngle + hornIncrUI) & DISPLC_MASK;
    drumAngle = (drumAngle + drumIncrUI) & DISPLC_MASK;

#undef HN_MOTION_A
#undef HN_MOTION_B
#undef DR_MOTION_A
#undef DR_MOTION_B
  }

  return outbuffer;
}

#ifdef WHIRL_MAIN
#define FAKEBUFSZ 256
int main (int argc, char * argv) {
  short inbuf [FAKEBUFSZ];
  short outbuf [2 * FAKEBUFSZ];
  int i;
  int iter;

  initWhirl ();
  setRevSelect (0);

  for (i = 0; i < DISPLC_SIZE; i++) {
    printf ("%4d %g\n", i, bfw[i][0]);
  }

  return 0;

  for (i = 0; i < FAKEBUFSZ; i++) {
    inbuf[i] = 10000;
  }
  for (iter = 0; iter < 2; iter++) {
    whirlProc ((unsigned char *) inbuf,
	       (unsigned char *) outbuf,
	       FAKEBUFSZ);
    printf ("BUFFER # %3d\n", iter);
    for (i = 0; i < FAKEBUFSZ; i++) {
      int j;
      printf ("%3d:%8d:%12d%12d\n", iter, i, outbuf[i * 2], outbuf[(i*2)+1]);
    }
  }
  
  return 0;
}

#endif /* WHIRL_MAIN */
